@@ -82,7 +82,7 @@ class Rows:
                 yield {**r1, **r2}
 
         # if the source is the database, it's safe to update rows directly
-        # because tablemap creates new rows everytime you iterate rows
+        # because tablemap creates new rows every time you iterate rows
         def gen2():
             for r1, r2 in zip(pgen(), _rows2iter(other)):
                 r1.update(r2)
@@ -185,6 +185,7 @@ class Rows:
 
         def gen_simp():
             xs = list(pgen())
+            # multiple column sorting needs an idea.
             for col in reversed(cols):
                 rev = False
                 cs = col.split()
@@ -383,17 +384,35 @@ class Rows:
 
         return self.apply(deselectfn)
 
-    def inner_join(self, other):
-        def innerjoin_fn(rs1, rs2, _left, _right):
+    def join(self, other, join_type="left"):
+        if join_type.lower() == "inner":
+            return self._inner_join(other)
+        elif join_type.lower() == "left":
+            return self._left_join(other)
+        elif join_type.lower() == "right":
+            return self._right_join(other)
+        elif join_type.lower() == "full":
+            return self._full_join(other)
+        raise ValueError("Unknown join_type", join_type)
+
+    def _inner_join(self, other):
+        def innerjoin_fn1(rs1, rs2, _left, _right_only):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     r1.update(r2)
                     yield r1
 
-        return self._merge(innerjoin_fn, other)
+        def innerjoin_fn2(rs1, rs2, _left, _right_only):
+            if rs1 != [] and rs2 != []:
+                for r1, r2 in product(rs1, rs2):
+                    yield {**r1, **r2}
 
-    def left_join(self, other):
-        def leftjoin_fn(rs1, rs2, _left, right):
+        if self._history[0]['cmd'] == 'fetch':
+            return self._merge(innerjoin_fn1, other)
+        return self._merge(innerjoin_fn2, other)
+
+    def _left_join(self, other):
+        def leftjoin_fn1(rs1, rs2, _left, right_only):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     r1.update(r2)
@@ -401,29 +420,50 @@ class Rows:
 
             elif rs1 != [] and rs2 == []:
                 for r1 in rs1:
-                    for col in right:
-                        r1[col] = ''
+                    r1.update((col, '') for col in right_only)
                     yield r1
 
-        return self._merge(leftjoin_fn, other)
-
-    def right_join(self, other):
-        def rightjoin_fn(rs1, rs2, left, _right):
+        def leftjoin_fn2(rs1, rs2, _left, right_only):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
-                    r2.update(r1)
-                    yield r2
+                    yield {**r1, **r2}
+
+            elif rs1 != [] and rs2 == []:
+                for r1 in rs1:
+                    yield {**r1, **{col: '' for col in right_only}}
+
+        if self._history[0]['cmd'] == 'fetch':
+            return self._merge(leftjoin_fn1, other)
+        return self._merge(leftjoin_fn2, other)
+
+    def _right_join(self, other):
+        def rightjoin_fn1(rs1, rs2, left, _right_only):
+            if rs1 != [] and rs2 != []:
+                for r1, r2 in product(rs1, rs2):
+                    r1.update(r2)
+                    yield r1
+
+            elif rs1 == [] and rs2 != []:
+                empty_left = {col: '' for col in left}
+                for r2 in rs2:
+                    empty_left.update(r2)
+                    yield empty_left
+
+        def rightjoin_fn2(rs1, rs2, left, _right_only):
+            if rs1 != [] and rs2 != []:
+                for r1, r2 in product(rs1, rs2):
+                    yield {**r1, **r2}
 
             elif rs1 == [] and rs2 != []:
                 for r2 in rs2:
-                    for col in left:
-                        r2[col] = ''
-                    yield r2
+                    yield {**{col: '' for col in left}, **r2}
 
-        return self._merge(rightjoin_fn, other)
+        if self._history[0]['cmd'] == 'fetch':
+            return self._merge(rightjoin_fn1, other)
+        return self._merge(rightjoin_fn2, other)
 
-    def full_join(self, other):
-        def fulljoin_fn(rs1, rs2, left, right):
+    def _full_join(self, other):
+        def fulljoin_fn1(rs1, rs2, left, right_only):
             if rs1 != [] and rs2 != []:
                 for r1, r2 in product(rs1, rs2):
                     # TODO: updating on the same object multiple times.
@@ -437,17 +477,31 @@ class Rows:
 
             elif rs1 != [] and rs2 == []:
                 for r1 in rs1:
-                    for col in right:
-                        r1[col] = ''
+                    r1.update((col, '') for col in right_only)
                     yield r1
 
             elif rs1 == [] and rs2 != []:
+                empty_left = {col: '' for col in left}
                 for r2 in rs2:
-                    d = {col: '' for col in left}
-                    d.update(r2)
-                    yield d
+                    empty_left.update(r2)
+                    yield empty_left
 
-        return self._merge(fulljoin_fn, other)
+        def fulljoin_fn2(rs1, rs2, left, right_only):
+            if rs1 != [] and rs2 != []:
+                for r1, r2 in product(rs1, rs2):
+                    yield {**r1, **r2}
+
+            elif rs1 != [] and rs2 == []:
+                for r1 in rs1:
+                    yield {**r1, **{col: '' for col in right_only}}
+
+            elif rs1 == [] and rs2 != []:
+                for r2 in rs2:
+                    yield {**{col: '' for col in left}, **r2}
+
+        if self._history[0]['cmd'] == 'fetch':
+            return self._merge(fulljoin_fn1, other)
+        return self._merge(fulljoin_fn2, other)
 
     def distinct(self, *columns):
         def distinctfn(_k, rs):
@@ -514,9 +568,20 @@ class Rows:
         }]
         return newself
 
+    def size(self):
+        if len(self._history) == 1 and self._history[0]['cmd'] == 'fetch':
+            origin = self._history[0]
+            return _get_size(origin['conn'], origin['tname'])
+        elif len(self._history) == 1 and self._history[0]['cmd'] == 'read':
+            return len(self._src)
+        return sum(1 for _ in self.iter())
+
+    # iter is not safe to end-users, because it can open the database.
+    # This should be used in a controlled way.
     def iter(self):
         yield from self._history[-1]['genfn']()
 
+    # list is safe, because it completes the iteration.
     def list(self):
         return list(self.iter())
 
@@ -593,38 +658,37 @@ def _step(fn, krs1, krs2):
         r1, rs1 = _spy(rs1)
         r2, rs2 = _spy(rs2)
 
-        left = [c for c in list(r1) if c not in list(r2)]
-        right = [c for c in list(r2) if c not in list(r1)]
+        left = list(r1)
+        right_only = [c for c in list(r2) if c not in list(r1)]
 
         while True:
             if k1 == k2:
-                yield from fn(rs1, rs2, left, right)
+                yield from fn(rs1, rs2, left, right_only)
                 k1 = k2 = empty
                 k1, rs1 = next(krs1)
                 k2, rs2 = next(krs2)
             elif k1 < k2:
-                yield from fn(rs1, [], left, right)
+                yield from fn(rs1, [], left, right_only)
                 k1 = empty
                 k1, rs1 = next(krs1)
             else:
-                yield from fn([], rs2, left, right)
+                yield from fn([], rs2, left, right_only)
                 k2 = empty
                 k2, rs2 = next(krs2)
 
     except StopIteration:
         # unconsumed
         if k1 is not empty:
-            yield from fn(rs1, [], left, right)
+            yield from fn(rs1, [], left, right_only)
         if k2 is not empty:
-            yield from fn([], rs2, left, right)
+            yield from fn([], rs2, left, right_only)
 
         for _, rs1 in krs1:
-            yield from fn(rs1, [], left, right)
+            yield from fn(rs1, [], left, right_only)
         for _, rs2 in krs2:
-            yield from fn([], rs2, left, right)
+            yield from fn([], rs2, left, right_only)
 
 
-# InstructionSet (It) to iter
 def _rows2iter(obj):
     return obj.iter() if isinstance(obj, Rows) else iter(obj)
 
@@ -641,6 +705,16 @@ def _build_initgen(conn, tname, cols):
     return initgen
 
 
+def _get_size(conn, tname):
+    try:
+        res = conn._dbconn.cursor().execute(f"select count(1) from {tname}") 
+        return res.fetchone()['count(1)']
+    except (AttributeError, sqlite3.ProgrammingError):
+        with _connect(conn._dbfile) as c:
+            res = c.cursor().execute(f"select count(1) from {tname}")
+            return res.fetchone()['count(1)']
+
+
 def _windowed(seq, n, step):
     if n < 0:
         raise ValueError('n must be >= 0')
