@@ -36,6 +36,25 @@ class TestCore(unittest.TestCase):
         if Path(self.dbfile1).is_file():
             os.remove(self.dbfile1)
 
+    def test_size(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1 
+
+        self.assertEqual(conn['t1'].size(), len(self.t1))
+        # size is executed while the connection is open
+        self.assertEqual(conn['t1'].update(n=lambda r: conn['t1'].size())["n"], 
+            [7] * 7)
+        # size is executed when the connection is closed
+        self.assertEqual(conn['t1'].update(n=conn['t1'].size())["n"], 
+            [7] * 7)
+
+        rs = Rows(self.t1)
+        self.assertEqual(rs.size(), len(self.t1))
+
+        rs1 = rs.islice(3,6)
+        self.assertEqual(rs1.size(), 3)
+         
+
     def test_wierd(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -47,6 +66,25 @@ class TestCore(unittest.TestCase):
                          {'col1': 'c', 'col2_tot': 26},
                          {'col1': 'd', 'col2_tot': 26}])
 
+    def test_modification(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        rs1 = conn['t1']
+        rs2 = rs1.update(col2=lambda r: r['col2'] + 1)
+
+        # the following code does not work, you cannot execute a fetch
+        # while another fetch is being done.
+        # So iter should not be exposed.
+        # for r1, r2 in zip(rs1.iter(), rs2.iter()):
+        #     print(r1, r2)
+
+        rs1 = rs1.list()
+        rs2 = rs2.list()
+
+        for r1, r2 in zip(rs1, rs2):
+            self.assertEqual(r1['col2'] + 1, r2['col2'])
+
     def test_list_mod(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -105,6 +143,13 @@ class TestCore(unittest.TestCase):
                           {'col1': 'd', 'col2': 2, 'n': 1}]
                          )
 
+    def test_zip2(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = self.t1
+
+        conn['t1_1'] = conn['t1'].zip({'idx': i} for i in range(100))
+        self.assertEqual(conn['t1_1']['idx'], list(range(len(self.t1))))
+
     def test_zip_longest(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -158,21 +203,21 @@ class TestCore(unittest.TestCase):
     def test_tee_n_new(self):
         conn = Conn(self.dbfile)
 
-        def tempfn(rs):
+        def tempfn1(rs):
             return rs.order('col2 desc').chain(rs.order('col2'))
         conn['t10'] = self.t1
 
-        conn['t1_1'] = conn['t10'].by('col1').map(tempfn)
+        conn['t1_1'] = conn['t10'].by('col1').map(tempfn1)
 
         self.assertEqual(conn['t1_1']['col2'],
                          [5, 4, 4, 5, 1, 1, 7, 4, 3, 3, 4, 7, 2, 2])
 
-        def tempfn(rs):
+        def tempfn2(rs):
             rs1 = rs.update(col2=lambda r: r['col2'] + 1)
             return rs1.order('col2 desc').chain(rs.order('col2'))
 
         conn['t1'] = self.t1
-        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn2)
         self.assertEqual(conn['t1_1']['col2'],
                          [6, 5, 4, 5, 2, 1, 8, 5, 4, 3, 4, 7, 3, 2])
 
@@ -463,19 +508,36 @@ class TestCore(unittest.TestCase):
         conn['t1'] = self.t1
         conn['t2'] = self.t2
 
-        conn['t1_1'] = conn['t1'].by('col1').inner_join(conn['t2'].by('col1'))
+        conn['t1_1'] = conn['t1'].by('col1')\
+            .join(conn['t2'].by('col1'), "inner")
 
         self.assertEqual(len(conn['t1_1'].list()), 3)
 
-        conn['t1_2'] = conn['t1'].by('col1').left_join(conn['t2'].by('col1'))
+        conn['t1_2'] = conn['t1'].by('col1')\
+            .join(conn['t2'].by('col1'), "left")
 
-        self.assertEqual(len(conn['t1_2'].list()), 8)
+        self.assertEqual(conn['t1_2'].list(),
+                         [{'col1': 'a', 'col2': 4, 'col3': ''},
+                          {'col1': 'a', 'col2': 5, 'col3': ''},
+                          {'col1': 'b', 'col2': 1, 'col3': 1},
+                          {'col1': 'c', 'col2': 3, 'col3': ''},
+                          {'col1': 'c', 'col2': 4, 'col3': ''},
+                          {'col1': 'c', 'col2': 7, 'col3': ''},
+                          {'col1': 'd', 'col2': 2, 'col3': 3},
+                          {'col1': 'd', 'col2': 2, 'col3': 2}])
 
-        conn['t1_3'] = conn['t1'].by('col1').right_join(conn['t2'].by('col1'))
+        conn['t1_3'] = conn['t1'].by('col1')\
+            .join(conn['t2'].by('col1'), "right")
 
-        self.assertEqual(len(conn['t1_3'].list()), 5)
+        self.assertEqual(conn['t1_3'].list(),
+                         [{'col1': 'b', 'col2': 1, 'col3': 1},
+                          {'col1': 'd', 'col2': 2, 'col3': 3},
+                          {'col1': 'd', 'col2': 2, 'col3': 2},
+                          {'col1': 'e', 'col2': '', 'col3': 3},
+                          {'col1': 'e', 'col2': '', 'col3': 4}])
 
-        conn['t1_4'] = conn['t1'].by('col1').full_join(conn['t2'].by('col1'))
+        conn['t1_4'] = conn['t1'].by('col1')\
+            .join(conn['t2'].by('col1'), "full")
 
         self.assertEqual(conn['t1_4'].list(), [
             {'col1': 'a', 'col2': 4, 'col3': ''},
@@ -497,15 +559,15 @@ class TestCore(unittest.TestCase):
 
         conn['t1_1'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
             'col1': 'd', 'col2': 19}]).by('col1')\
-            .full_join(conn['t2'].by('col1'))
+            .join(conn['t2'].by('col1'), 'full')
 
         conn['t1_1'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
             'col1': 'd', 'col2': 19}]).by('col1')\
-            .full_join(conn['t2'].by('col1'))
+            .join(conn['t2'].by('col1'), 'full')
 
         conn['t1_2'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
             'col1': 'd', 'col2': 19}]).by('col1')\
-            .full_join(conn['t2'].by('col1'))\
+            .join(conn['t2'].by('col1'), 'full')\
             .update(col4=lambda r: r['col3'] + 1 if r['col3'] else '')
 
         self.assertEqual(conn['t1_1'].list(),
@@ -524,6 +586,22 @@ class TestCore(unittest.TestCase):
                           {'col1': 'e', 'col2': '', 'col3': 3},
                           {'col1': 'e', 'col2': '', 'col3': 4}])
 
+    def test_join3(self):
+        rs1 = Rows(self.t1)
+        rs2 = Rows(self.t2)
+        rs = rs1.by('col1').join(rs2.by('col1'))
+        self.assertEqual(rs.list(),
+                         [{'col1': 'a', 'col2': 4, 'col3': ''},
+                          {'col1': 'a', 'col2': 5, 'col3': ''},
+                          {'col1': 'b', 'col2': 1, 'col3': 1},
+                          {'col1': 'c', 'col2': 3, 'col3': ''},
+                          {'col1': 'c', 'col2': 4, 'col3': ''},
+                          {'col1': 'c', 'col2': 7, 'col3': ''},
+                          {'col1': 'd', 'col2': 2, 'col3': 3},
+                          {'col1': 'd', 'col2': 2, 'col3': 2}])
+
+        self.assertEqual(rs1.list(), self.t1)
+
     def test_distinct(self):
         conn = Conn(self.dbfile)
         conn['t1'] = self.t1
@@ -620,7 +698,7 @@ class TestCore(unittest.TestCase):
         ]
 
         conn['t1_col3'] = conn['t1'].by('col1')\
-            .full_join(conn['t2'].by('col1'))
+            .join(conn['t2'].by('col1'), 'full')
 
         self.assertEqual(conn['t1_col3'].list(),
                          [{'col1': 'a', 'col2': 4, 'col3': ''},
